class BlockInfo:
    """Details about a specific SOR block"""

    identifier: str
    revision_number: int
    size: int

class MapBlock:
    """File layout information.
    Every SOR file has a MapBlock which acts as a map to the file's contents"""

    revision_number: int
    block_size: int
    block_count: int
    block_info: list[BlockInfo]

class GeneralParametersBlock:
    """General details about this trace"""

    language_code: str
    """Language code - EN, CN, JP, etc."""
    cable_id: str
    """Cable identifier"""
    fiber_id: str
    """Fibre identifier"""
    fiber_type: int
    """Fibre type - this is generally coded as the ITU-T standard definition"""
    nominal_wavelength: int
    """Nominal test wavelength in nm"""
    originating_location: str
    """Start location for the test"""
    terminating_location: str
    """End location for the test"""
    cable_code: str
    """Cable code - free field"""
    current_data_flag: str
    """ NC for new condition, RC for as-repaired, OT as something else"""
    user_offset: int
    """User offset - This is essentially the launch lead length from the front panel offset (provided in the fixed parameters block), in 100ps increments"""
    user_offset_distance: int
    """This is the same as user_offset, but measured in 10x the distance units of FixedParametersBlock.units_of_distance"""
    operator: str
    """Operator of the unit for the test"""
    comment: str
    """Free comment field"""

class SupplierParametersBlock:
    """Details about the OTDR used in the test"""

    supplier_name: str
    """Manufacturer of the OTDR"""
    otdr_mainframe_id: str
    """Mainframe model number"""
    otdr_mainframe_sn: str
    """Mainframe serial number"""
    optical_module_id: str
    """Optical module model number"""
    optical_module_sn: str
    """Optical module serial number"""
    software_revision: str
    """Software revision"""
    other: str
    """Free text"""

class FixedParametersBlock:
    """Details about the trace that don't vary between datasets"""

    date_time_stamp: int
    """Datestamp - unix epoch seconds, 32-bit. Remember not to do any OTDR tests after 2038."""
    units_of_distance: str
    """Units of distance - km, mt, ft, kf, mi, etc. Typically mt (in civilised nations)"""
    actual_wavelength: int
    """Actual wavelength used - normally the factory-calibrated wavelength in nm, or nominal wavelength"""
    acquisition_offset: int
    """Acquisition offset - this is the length of fibre from the OTDR port to the first data point in the DataPoints, in 100ps increments"""
    acquisition_offset_distance: int
    """As acquisition_offset, but as 10x units_of_distance"""
    total_n_pulse_widths_used: int
    """The total number of pulse widths used, if more than one pulse width's results are stored in the file"""
    pulse_widths_used: list[int]
    """The list of pulse widths used, in nanoseconds"""
    data_spacing: list[int]
    """Data spacing, time taken to acquire 10,000 points in 100ps increments"""
    n_data_points_for_pulse_widths_used: list[int]
    """Number of points stored for each pulse width"""
    group_index: int
    """Group index - the refractive index of the fibre, default to 146800 if nothing supplied"""
    backscatter_coefficient: int
    """Backscatter coefficient -"""
    number_of_averages: int
    """Number of averages - the number of samples that were averaged to generate the result - may be used instead of averaging time"""
    averaging_time: int
    """Averaging time - may be supplied instead of number of averages - in seconds x 10"""
    acquisition_range: int
    """Acquisition range set by the tester to reach the end of the fibre - as with other distance measurements, 100ps increments"""
    acquisition_range_distance: int
    """Acquisition range in 10x distance units, as an alternate to acquisition_range"""
    front_panel_offset: int
    """Front panel offset is the time taken, in 100ps increments, between the front-end of the optical TRX and the front panel connector"""
    noise_floor_level: int
    """Noise floor level - the lowest power level for which 98% of the noise data lies below; 5-digit -dB value (e.g. 10200 = -10.2dB)"""
    noise_floor_scale_factor: int
    """Scale factor for the noise floor level - defaults to 1"""
    power_offset_first_point: int
    """Attenuation in dB*1000 applied by the instrument if done by the instrument"""
    loss_threshold: int
    """The threshold in dB*1000 for a loss-type event; default 00200"""
    reflectance_threshold: int
    """The threshold in -dB*1000 for reflectance events; default -55000"""
    end_of_fibre_threshold: int
    """The threshold in dB*1000 for the loss taken to detect the end of the fibre; default 03000"""
    trace_type: str
    """Trace type - identifies if this is a standard one-way trace, a bidirectional trace, reference trace, difference trace, or reversed trace"""
    window_coordinate_1: int
    """Window coordinate for the upper right window corner"""
    window_coordinate_2: int
    """Power coordinate for the upper right window corner"""
    window_coordinate_3: int
    """Window coordinate for the lower left window corner"""
    window_coordinate_4: int
    """Power coordinate for the lower left window corner"""

class KeyEvent:
    """Event detected by the OTDR"""

    event_number: int
    """Event number - this is from 0 to n"""

    event_propogation_time: int
    """Event propogation time is the time in 100ps units from the front panel to the event"""

    attenuation_coefficient_lead_in_fiber: int
    """The span loss in db/km (as a 5-digit value, i.e. dB*1000) for the fibre entering the event"""

    event_loss: int
    """Loss in dB*1000 for the event"""

    event_reflectance: int
    """Reflectance in -dB*1000 for the event"""

    event_code: str
    """Code for the event is a 6-byte string:
    Byte 1:
        0 = nonreflective, 1 = reflective, 2 = saturated reflective
    Byte 2:
        A = added by user, M = moved by user, E = end of fibre, F = found
        by software, O = out of range, D = modified end of fibre
    Remaining bytes are the Landmark number if used - 9s otherwise
    """

    loss_measurement_technique: str
    """Loss measurement technique - 2P for two point, LS for least squares, OT for other"""

    marker_location_1: int
    """Marker location - ML1 is the OTDR side for 2P/LS/OT measurements"""

    marker_location_2: int
    """Marker location - ML2 is the OTDR side for LS measurements, and bounds the event for 2P/OT"""

    marker_location_3: int
    """Marker location - ML3 is on the far side for LS measurements, and empty for 2P/OT"""

    marker_location_4: int
    """Marker location - ML4 is on the far side for LS measurements, and empty for 2P/OT"""

    marker_location_5: int
    """Marker location - ML5 is the reflectance calculation position"""

    comment: str
    """Free comment on the event"""

class LastKeyEvent:
    """Like a KeyEvent, but specific to the last event detected"""

    event_number: int
    event_propogation_time: int
    attenuation_coefficient_lead_in_fiber: int
    event_loss: int
    event_reflectance: int
    event_code: str
    loss_measurement_technique: str
    marker_location_1: int
    marker_location_2: int
    marker_location_3: int
    marker_location_4: int
    marker_location_5: int
    comment: str
    # End to end loss is in dB*1000 and measures the loss between the two
    # markers defined below
    end_to_end_loss: int
    # Start of the measurement span - typically user offset
    end_to_end_marker_position_1: int
    # End of the measurement span - typically end of fibre event position
    end_to_end_marker_position_2: int
    # Return loss in dB*1000 for the markers defined below
    optical_return_loss: int
    # Start of the measurement span - typically user offset
    optical_return_loss_marker_position_1: int
    # End of the measurement span - typically end of fibre event position
    optical_return_loss_marker_position_2: int

class KeyEvents:
    """The key events the OTDR detected."""

    number_of_key_events: int
    key_events: list[KeyEvent]
    last_key_event: LastKeyEvent

class DataPointsAtScaleFactor:
    """The data points for a specific scale factor"""

    n_points: int
    """Number of points in this block"""
    scale_factor: int
    """Scale factor for the data, as 1000*SF"""
    data: list[int]
    """Data points as dB*1000"""

class DataPoints:
    """The data points for this trace"""

    number_of_data_points: int
    total_number_scale_factors_used: int
    scale_factors: list[DataPointsAtScaleFactor]

class Landmark:
    """Not widely used; a landmark recorded in the trace by the system."""

    landmark_number: int
    # Landmark code identifies the landmark - see page 27 of the standard for
    # the list
    landmark_code: str
    # Location in 100ps from user offset to the landmark
    landmark_location: int
    related_event_number: int
    gps_longitude: int
    gps_latitude: int
    # Fibre correction factor is the difference in 100*% between the optical
    # path and the cable length; otherwise known as heliax correction
    fiber_correction_factor_lead_in_fiber: int
    sheath_marker_entering_landmark: int
    sheath_marker_leaving_landmark: int
    units_of_sheath_marks_leaving_landmark: str
    mode_field_diameter_leaving_landmark: int
    comment: str

class LinkParameters:
    """Landmark information for the trace"""

    number_of_landmarks: int
    landmarks: list[Landmark]

class ProprietaryBlock:
    """Binary proprietary block data."""

    header: str
    data: list[int]

class SORFile:
    """A SOR file.
    SORFile describes a full SOR file. All blocks except MapBlock are optional types as we cannot
    guarantee the parser will find them, but many blocks are in fact mandatory in the specification,
    so compliant instruments will provide them.
    """

    map: MapBlock
    general_parameters: GeneralParametersBlock | None
    supplier_parameters: SupplierParametersBlock | None
    fixed_parameters: FixedParametersBlock | None
    key_events: KeyEvents | None
    link_parameters: LinkParameters | None
    data_points: DataPoints | None
    proprietary_blocks: list[ProprietaryBlock]

def parse_file(path: str) -> SORFile:
    """Load a SOR from the given path and parse it"""

def parse_bytes(bytes: bytes) -> SORFile:
    """Parse a SOR file from the given bytes"""



#!/bin/env python3
# Simple demo script to call otdrs, parse out a few variables, and draw a data point graph with the front panel marked as a line with everything scaled appropriately
# Assumes metres are the units used in the SOR file.
import numpy as np

# You can use CBOR instead of JSON by swapping this and adding the --format cbor option below
# import cbor
import json
import subprocess
import otdrs
import matplotlib.pyplot as plt

sor = otdrs.parse_file("data/example1-noyes-ofl280.sor")

if (
    sor.fixed_parameters is None
    or sor.data_points is None
    or sor.general_parameters is None
    or sor.key_events is None
):
    raise RuntimeError("Missing some key blocks in this file!")

# To properly scale distances, we need to calculate the speed of light in fibre. This varies from fibre to fibre, but is about 1.47ish.
speed_of_light = 299792458.0  # m/s
# To accommodate this variation, the group index may be stored in the data (though if this is 0 you should use 146800)
refractive_index = sor.fixed_parameters.group_index / 100000.0
speed_of_light_in_fibre = (
    speed_of_light / refractive_index
)  # Calculate our speed of light for the reported group index/refractive index

# The following assumes only one pulse width used - this is true for almost all OTDR traces.
# Even acquisition regimes such as EXFO iOLM don't use the multiple pulse width/spacing convention, and just store data in proprietary blobs.
seconds_per_10k_points = (
    sor.fixed_parameters.data_spacing[0] / 1e10
)  # convert from picoseconds up
metres_per_data_spacing = (
    seconds_per_10k_points / 10000.0
) * speed_of_light_in_fibre  # FIXME: actually check this maths is right!
# Assumes only one scale factor used - same as above, this is generally "safe"!
sf = sor.data_points.scale_factors[0].scale_factor  # multiplier for the data
scaled_data = -(65535 - np.array(sor.data_points.scale_factors[0].data)) / float(
    sf
)  # Apply the scale factor to the whole dataset
# Just for cosmetics (in this example) we'll draw lines at the front panel point and the launch connector point
seconds_to_front_panel = sor.fixed_parameters.front_panel_offset / 1e10
seconds_to_launch_connector = sor.general_parameters.user_offset / 1e10
# And in metres, that's distance = time * speed
metres_to_front_panel = seconds_to_front_panel * speed_of_light_in_fibre
# Same again for launch - but we do need to offset from the front panel...
metres_to_launch_connector = (
    seconds_to_launch_connector * speed_of_light_in_fibre
) + metres_to_front_panel


# Let's print out our key events - helper function to avoid duplication for last_key_event...
def print_key_event(ke, sf, sol, fpo):
    loss = ke.event_loss / sf
    reflectance = ke.event_reflectance / sf
    seconds_to_event = ke.event_propogation_time / 1e10
    metres_to_event = (seconds_to_event * sol) + fpo
    print(
        "Event {}: {}dB loss, {}dB reflectance, {}m".format(
            ke.event_number, loss, reflectance, metres_to_event
        )
    )


# Now print the lot
for ke in sor.key_events.key_events:
    print_key_event(ke, sf, speed_of_light_in_fibre, metres_to_front_panel)
print_key_event(
    sor.key_events.last_key_event, sf, speed_of_light_in_fibre, metres_to_front_panel
)

print("E2E loss {}dB".format(sor.key_events.last_key_event.end_to_end_loss / sf))

# Plot our data - we calculate our X axis values based on the spacing interval and number of points to plot, on our scale of metres
# Note that "0" will be from the first data point, behind the front panel - normally we'd want to offset everything
# so that the 0 point is the launch connector, since that makes our subsequent event data make sense!
spacing = np.arange(
    0,
    metres_per_data_spacing * sor.data_points.scale_factors[0].n_points,
    metres_per_data_spacing,
)[0 : sor.data_points.scale_factors[0].n_points]
plt.plot(
    spacing,
    scaled_data,
    linewidth=1,
)
plt.axvline(x=metres_to_front_panel)  # draw a line at the front panel
plt.axvline(x=metres_to_launch_connector)  # draw a line at the front panel
plt.xlabel("Metres from OTDR module (not front panel/launch)")
plt.ylabel("dB")

plt.gca().invert_yaxis()
plt.show()
